use std::fs;

use anyhow::{Error, Result};
use cli_log::{error, info};
use matrix_sdk::{ruma::exports::serde_json, Client, Session};
use serde::{Deserialize, Serialize};

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct Account {
    homeserver: String,
    id: u32,
    name: String,

    session: Session,
    sync_token: Option<String>,
}

#[derive(Debug, Serialize, Deserialize)]
struct AccountsData {
    current_account: u32,
    accounts: Vec<Account>,
}

pub struct AccountsManager {
    current_account: u32,
    num_accounts: u32,
    accounts: Vec<Account>,
    clients: Vec<Option<Client>>,
}

impl Account {
    pub fn name(&self) -> &String {
        &self.name
    }

    pub fn user_id(&self) -> String {
        self.session.user_id.to_string()
    }
}

impl AccountsManager {
    pub fn new(config: Option<String>) -> Result<Self> {
        return match config {
            Some(s) => {
                info!("Loading serialized AccountsManager");
                let accounts_data: AccountsData = serde_json::from_str(&s)?;
                let mut clients = Vec::new();
                clients.resize(accounts_data.accounts.len(), None);
                Ok(Self {
                    current_account: accounts_data.current_account,
                    num_accounts: accounts_data.accounts.len() as u32,
                    accounts: accounts_data.accounts,
                    clients,
                })
            }
            None => {
                info!("Creating empty AccountsManager");
                Ok(Self {
                    current_account: 0,
                    num_accounts: 0,
                    accounts: Vec::new(),
                    clients: Vec::new(),
                })
            }
        };
    }

    pub async fn restore(&mut self) -> Result<()> {
        self.login(self.current_account).await?;
        Ok(())
    }

    pub async fn add(
        &mut self,
        homeserver: &String,
        username: &String,
        password: &String,
    ) -> Result<u32> {
        let id = self.num_accounts;
        self.num_accounts += 1;

        let client = Client::builder()
            .homeserver_url(homeserver)
            .sled_store(format!("userdata/{id}"), Some("supersecure"))?
            .build()
            .await?;

        client
            .login_username(username, password)
            .initial_device_display_name("Trinitrix")
            .send()
            .await?;

        let session = match client.session() {
            Some(s) => s,
            None => return Err(Error::msg("Failed to get session")),
        };

        let name = match client.account().get_display_name().await? {
            Some(n) => n,
            None => return Err(Error::msg("Failed to get display name")),
        };

        let account = Account {
            homeserver: homeserver.to_string(),
            id,
            name,
            session: session.clone(),
            sync_token: None,
        };

        self.logout().await?;
        self.current_account = id;
        self.accounts.push(account);
        self.clients.push(Some(client));
        self.save()?;

        info!(
            "Logged in as '{}' device ID: {}",
            session.user_id.to_string(),
            session.device_id.to_string()
        );

        Ok(id)
    }

    pub async fn login(&mut self, account_id: u32) -> Result<()> {
        self.logout().await?; // log out the current account

        let account = if account_id >= self.num_accounts {
            error!("Tried to log in with an invalid account ID {}", account_id);
            return Err(Error::msg("Invalid account ID"));
        } else {
            if let Some(a) = self.get(account_id) {
                a
            } else {
                return Err(Error::msg("Failed to get account"));
            }
        };

        if self
            .clients
            .get(account_id as usize)
            .expect("client lookup failed")
            .is_none()
        {
            info!(
                "No client cached for account: '{}' -> requesting a new one",
                &account.session.user_id
            );
            let client = Client::builder()
                .homeserver_url(&account.homeserver)
                .sled_store(format!("userdata/{account_id}"), Some("supersecure"))?
                .build()
                .await?;
            client.restore_login(account.session.clone()).await?;
            self.clients.insert(account_id as usize, Some(client));
        } else {
            info!(
                "Using cached client for account: '{}'",
                &account.session.user_id
            );
        };

        info!("Restored account");

        self.current_account = account_id;

        Ok(())
    }

    pub async fn logout(&mut self) -> Result<()> {
        // idk, do some matrix related stuff in here or something like that
        if self.clients.get(self.current_account as usize).is_none() {
            return Ok(());
        }
        let client = match self.clients.get(self.current_account as usize).unwrap() {
            None => return Ok(()),
            Some(c) => c,
        };

        info!("Logged out '{}' locally", client.session().unwrap().user_id);
        client.logout().await?;

        Ok(())
    }

    pub fn save(&self) -> Result<()> {
        let accounts_data = AccountsData {
            current_account: self.current_account,
            accounts: self.accounts.clone(),
        };

        let serialized = serde_json::to_string(&accounts_data)?;
        fs::write("userdata/accounts.json", serialized)?;

        info!("Saved serialized accounts config (userdata/accounts.json)");

        Ok(())
    }

    pub fn get(&self, id: u32) -> Option<&Account> {
        self.accounts.get(id as usize)
    }

    pub fn current(&self) -> Option<&Account> {
        self.get(self.current_account)
    }

    pub fn client(&self) -> Option<&Client> {
        match self.clients.get(self.current_account as usize) {
            None => None,
            Some(oc) => match oc {
                None => None,
                Some(c) => Some(c),
            },
        }
    }

    pub fn num_accounts(&self) -> u32 {
        self.num_accounts
    }
}
