// Run the `api` bin to see the generated api

use std::{panic::catch_unwind, process, thread};

use cli_log::{debug, info};

use crate::app::{events::Event, COMMAND_TRANSMITTER};

include!(concat!(env!("OUT_DIR"), "/api.rs"));

pub fn handle_cmd(cmd: Commands) {
    // Unwinding into the c code implicitly calling this function would be UB. Besides, rust would
    // probably just abort the process, giving us a weird shutdown without an error message.
    // Thus we catch the panic before the ffi boundary. This could also be moved to trixy.
    let error = catch_unwind(|| {
        let tx = COMMAND_TRANSMITTER
            .get()
            .expect("The cell should always be populated, at this point");

        info!("Received command: {:#?}", cmd);

        // NOTE: The extra future and tokio runtime here is necessary to have a sync api. Which is imho
        // better than expecting c to work with a async one. <2024-05-04>
        let future = async move {
            debug!("Asyncly started to send cmd");

            // FIXME: The None here is definitely wrong <2024-05-03>
            tx.send(Event::CommandEvent(cmd, None))
                .await
                .expect("I hope that this does not fail");

            debug!("Done with the async send");
        };

        debug!("Starting runtime");
        let handle = thread::spawn(|| {
            // run the task a separate thread to avoid tokio having problems with the currently
            // running runtime.

            // PERFORMANCE(@soispha): Stating a new thread for each command is very bad. <2024-05-04>
            tokio::runtime::Builder::new_current_thread()
                .enable_all()
                .build()
                .unwrap()
                .block_on(future);
        });
        handle.join().expect("Shall never error");
        debug!("Handled command");
    });

    if let Err(err) = error {
        eprintln!("Catched a panic just before the c ffi: {:#?}", err);

        // This will leave the terminal in a horrendous shape (as no destructors are run), but what
        // can we do at this point beside that?
        process::abort();
    }
}
