use std::collections::HashMap;

use cli_log::info;
use mlua::{ErrorContext, FromLua, IntoLua, LuaSerdeExt, Value};

use super::{support_types::Function, CommandTransferValue, Table};

impl<'lua> FromLua<'lua> for Function {
    fn from_lua(value: Value<'lua>, lua: &'lua mlua::Lua) -> mlua::Result<Self> {
        match value {
            Value::Integer(function_id) => {
                return Ok(Function::new(function_id.try_into().expect(
                    "We should never have i64::MAX functions
                     registered. The stack will
                     probably overflow, before this i64 overflows",
                )))
            }
            _ => unreachable!("The Function type can only take functions!"),
        };
    }
}

impl<'lua> IntoLua<'lua> for CommandTransferValue {
    fn into_lua(self, lua: &'lua mlua::Lua) -> mlua::Result<mlua::Value<'lua>> {
        let converted_output = lua.to_value(&self)?;
        return unwrap(converted_output, lua);
    }
}
impl<'lua> FromLua<'lua> for CommandTransferValue {
    fn from_lua(value: Value<'lua>, lua: &'lua mlua::Lua) -> mlua::Result<Self> {
        match value {
            Value::Nil => Ok(Self::Nil),
            Value::Boolean(bool) => Ok(Self::Boolean(bool)),
            Value::Integer(int) => Ok(Self::Integer(int)),
            Value::Number(num) => Ok(Self::Number(num)),
            Value::String(string) => Ok(Self::String(
                string
                    .to_str()
                    .context("Lua string seems to contain non UTF-8 chars")?
                    .to_owned(),
            )),
            Value::Table(table) => {
                // FIXME(@soispha): This will fail, when the key is not a string, this should be
                // accounted for <2023-09-09>
                let mut rust_table: Table = HashMap::new();
                for pair in table.pairs() {
                    let (key, value) = pair?;
                    rust_table.insert(key, value);
                }
                return Ok(Self::Table(rust_table));
            }

            Value::Function(_) => todo!(),
            Value::Thread(_) => todo!(),
            Value::UserData(_) => todo!(),
            Value::Error(_) => todo!(),
            Value::LightUserData(_) => todo!(),
        }
    }
}

fn unwrap<'lua>(
    value_to_unwrap: Value<'lua>,
    lua: &'lua mlua::Lua,
) -> mlua::Result<mlua::Value<'lua>> {
    fn unwrap_first_level<'lua>(table: mlua::Table<'lua>) -> mlua::Result<Value<'lua>> {
        let (_, value): (Value, Value) = table
            .pairs()
            .next()
            .expect("Exactly one item should extist")?;
        Ok(value)
    }
    // That converted value looks somewhat like this (e.g. a String):
    // ```
    // {
    //   ["String"] = "hi",
    // }
    // ```
    // or like this (e.g. a table):
    // ```
    // {
    //   ["Table"] = {
    //     ["UserId"] = {
    //       ["Integer"] = 2,
    //     },
    //     ["UserName"] = {
    //       ["String"] = "James",
    //     },
    //   },
    // }
    // ```
    if let Value::Table(table) = value_to_unwrap {
        let value = unwrap_first_level(table)?;
        if let Value::Table(wrapped_table) = value {
            info!("We've got a wtable! wtable: \n{:#?}", wrapped_table);
            // we now have a wrapped table value for example like this:
            // ```
            // {
            //   ["UserId"] = {
            //     ["Integer"] = 2,
            //   },
            //   ["UserName"] = {
            //     ["String"] = "James",
            //   },
            //   ["Versions"] = {
            //     ["Table"] = {
            //       ["api"] = {
            //         ["Boolean"] = true,
            //       },
            //       ["interface"] = {
            //         ["Integer"] = 3,
            //       },
            //     },
            //   },
            // }
            // ```
            let output_table: mlua::Table = lua
                .load("{}")
                .eval()
                .expect("This is static, it should always work");

            // FIXME(@soispha): This still fails for nested tables (i.e. the table above), as it
            // unpacks too much. While unpacking the while loop should stop, when a key is not from
            // the CommandTransferValue family (i.e. ["Integer", "Boolean", "String", "Table",
            // etc.]) <2023-09-09>
            for pair in wrapped_table.pairs::<Value, Value>() {
                let (key, mut raw_value) = pair?;
                while let Value::Table(raw_table) = raw_value {
                    raw_value = unwrap_first_level(raw_table)?;
                }
                output_table.set(key, raw_value)?;
            }

            return Ok(output_table.into_lua(lua)?);
        } else {
            info!("We've got a normal output! output: {:#?}", value);
            // we had a simple wrapped value, which is already unwrapped, thus it can be
            // returned directly
            return Ok(value);
        }
    } else {
        unreachable!("The returned table should always only contain one element");
    }
}
