use std::{collections::HashMap, fmt::Display};

use serde::{Deserialize, Serialize};

use self::support_types::Function;

pub mod support_types;
pub mod type_conversions;

// language support
pub mod lua;

pub type Table = HashMap<String, CommandTransferValue>;

#[derive(Debug, Serialize, Deserialize, Clone)]
pub enum CommandTransferValue {
    /// `nil` or `null` or `undefined`; anything which goes in that group of types.
    Nil,

    /// `true` or `false`.
    Boolean(bool),

    // A “light userdata” object, equivalent to a raw pointer.
    // /* TODO */ LightUserData(LightUserData),
    /// An integer number.
    Integer(i64),

    /// A floating point number.
    Number(f64),

    /// A string
    String(String),

    /// A table, dictionary or HashMap
    Table(HashMap<String, CommandTransferValue>),

    /// Reference to a function (or closure).
    /// This 'Function' value is obtained by registering a function with 'register_function()'
    Function(Function),
    // Reference to a Lua thread (or coroutine).
    // /* TODO */ Thread(Thread<'lua>),

    // Reference to an userdata object that holds a custom type which implements `UserData`.
    // Special builtin userdata types will be represented as other `Value` variants.
    // /* TODO */ UserData(AnyUserData),

    // `Error` is a special builtin userdata type. When received from Lua it is implicitly cloned.
    // /* TODO */ Error(Error),
}

impl Display for CommandTransferValue {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            CommandTransferValue::Nil => f.write_str("Nil"),
            CommandTransferValue::Boolean(bool) => f.write_str(&format!("{}", bool)),
            CommandTransferValue::Integer(int) => f.write_str(&format!("{}", int)),
            CommandTransferValue::Number(num) => f.write_str(&format!("{}", num)),
            CommandTransferValue::String(str) => f.write_str(&format!("{}", str)),
            // TODO(@Soispha): The following line should be a real display call, but how do you
            // format a HashMap?
            CommandTransferValue::Table(table) => f.write_str(&format!("{:#?}", table)),
            CommandTransferValue::Function(function) => function.fmt(f),
        }
    }
}
