use std::{
    fmt::Display,
    sync::atomic::{AtomicUsize, Ordering},
};

use cli_log::info;
use mlua::{ErrorContext, IntoLua, Lua, Table};
use serde::{Deserialize, Serialize};

#[derive(Clone, Copy, Debug, Deserialize, Serialize, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub struct Function {
    id: usize,
}

impl Function {
    fn get_private<'lua>(lua: &'lua Lua) -> mlua::Result<Table<'lua>> {
        let private: Table = lua
            .globals()
            // This is always initialized, as the namespaces are specified in the 'command_list' module
            .get::<&str, mlua::Table>("trinitrix")
            .context("Failed to access 'trinitrix'")?
            .get::<&str, mlua::Table>("api")
            .context("Failed to access 'api'")?
            .get::<&str, mlua::Table>("raw")
            .context("Failed to access 'raw'")?
            .get::<&str, mlua::Table>("__private")
            .context("Failed to access '__private'")?;
        Ok(private)
    }
    pub fn new(function_id: usize) -> Self {
        Function { id: function_id }
    }
    pub fn from_lua_function(function: mlua::Function, lua: &Lua) -> mlua::Result<Self> {
        // TODO(@soispha): Does this expose a vulnerability, as the ids are predictable? <2023-10-14>
        static COUNTER: AtomicUsize = AtomicUsize::new(0);
        let id = COUNTER.fetch_add(1, Ordering::Relaxed);

        let private = Self::get_private(lua)?;
        info!("Registering function '{}'", id);
        private.set(id, function)?;

        Ok(Function::new(id))
    }
    pub fn call(&self, lua: &Lua) -> mlua::Result<()> {
        let private = Self::get_private(lua)?;

        info!("Calling function '{}'", &self.id);

        let function: mlua::Function = private
            .get(self.id)
            .context("Failed to get function associated with callback!")?;

        function.call(()).context("Failed to call function")
    }
}

impl<'lua> IntoLua<'lua> for Function {
    fn into_lua(self, lua: &'lua Lua) -> mlua::Result<mlua::Value<'lua>> {
        Ok(self.id.into_lua(lua)?)
    }
}

impl Display for Function {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.id.fmt(f)
    }
}
