use std::thread;

use anyhow::{Context, Result};
use cli_log::{debug, error, info};
use mlua::{ErrorContext, Lua, Value};
use once_cell::sync::OnceCell;
use tokio::{
    runtime::Builder,
    select,
    sync::{mpsc, Mutex},
    task::{self, LocalSet},
};

use crate::app::{
    command_interface::{
        add_lua_functions_to_globals,
        Api::Raw,
        Command,
        Raw::{DisplayOutput, RaiseError},
        Trinitrix::Api,
    },
    events::Event,
};

use super::command_transfer_value::support_types::Function;

static LUA: OnceCell<Mutex<mlua::Lua>> = OnceCell::new();

/// This structure contains the necessary state for running an embedded Lua runtime (i.e.
/// the tread, the Lua memory, etc.).
pub struct LuaCommandManager {
    lua_command_tx: mpsc::Sender<String>,
    lua_function_tx: mpsc::Sender<Function>,
}

impl LuaCommandManager {
    pub async fn execute_code(&self, code: String) {
        self.lua_command_tx
            .send(code)
            .await
            .expect("The receiver should not be dropped at this time");
    }
    pub async fn execute_function(&self, function: Function) {
        self.lua_function_tx
            .send(function)
            .await
            .expect("The receiver should not be dropped");
    }

    pub fn new(event_call_tx: mpsc::Sender<Event>) -> Self {
        info!("Spawning lua code execution thread...");
        let (lua_command_tx, mut lua_command_rx) = mpsc::channel::<String>(256);
        let (lua_function_tx, mut lua_function_rx) = mpsc::channel::<Function>(256);
        thread::spawn(move || {
            let rt = Builder::new_current_thread().enable_all().build().expect(
                "Should always be able to build \
                tokio runtime for lua command handling",
            );
            let local = LocalSet::new();
            local.spawn_local(async move {
                info!(
                    "Lua command handling initialized, \
                waiting for commands.."
                );
                let mut done = false;
                let moved_event_call_tx = event_call_tx.clone();
                while !done {
                    select! {
                        command = lua_command_rx.recv() => {
                                if let Some(command) = command {
                                        debug!("Received lua code (in LuaCommandHandler): {}", &command);
                                        let local_event_call_tx = moved_event_call_tx.clone();

                                        task::spawn_local(async move {
                                            exec_lua(&command, local_event_call_tx).await.expect(
                                                "This should return all relevent errors \
                                                     by other messages, \
                                                     this should never error",
                                            );
                                        });
                                } else {
                                    done = true;
                                }
                            },
                        function_id = lua_function_rx.recv() => {
                                if let Some(function_id) = function_id {
                                    debug!("Received lua function (in LuaCommandHandler): {}", &function_id);
                                    let local_event_call_tx = moved_event_call_tx.clone();

                                    task::spawn_local(async move {
                                         let lua = initialize_lua(local_event_call_tx.clone()).await;
                                         let out = function_id.call(&lua).map_err(|err| async move {
                                              error!("Lua function `{}` returned error: `{}`", function_id, err);
                                              local_event_call_tx
                                                  .send(Event::CommandEvent(
                                                      Command::Trinitrix(Api(Raw(RaiseError(err.to_string())))),
                                                      None,
                                                  ))
                                                  .await.expect(
                                                    "This should return all relevent errors \
                                                     by other messages, \
                                                     this should never error",
                                                    );
                                             });
                                          if let Err(err) = out {
                                              err.await;
                                          }
                                    });
                                } else {
                                    done = true;
                                }
                        },
                        else => done = true,
                    }
                }
            });
            rt.block_on(local);
        });

        LuaCommandManager {
            lua_command_tx,
            lua_function_tx,
        }
    }
}

async fn initialize_lua<'a>(
    event_call_tx: mpsc::Sender<Event>,
) -> matrix_sdk::locks::MutexGuard<'a, Lua> {
    let lua = LUA
        .get_or_init(|| {
            let lua: Lua = add_lua_functions_to_globals(mlua::Lua::new(), event_call_tx);
            {
                let wrapped_register_function = lua
                    .create_function(
                        |lua: &Lua, function: mlua::Function| -> mlua::Result<Function> {
                            ErrorContext::context(
                                Function::from_lua_function(function, lua),
                                "Failed to register function",
                            )
                        },
                    )
                    .expect("This always works, as the function is static");
                let trinitrix_api: mlua::Table = lua
                    .globals()
                    .get::<&str, mlua::Table>("trinitrix")
                    .expect("This was set in the add_lua_functions_to_globals function")
                    .get::<&str, mlua::Table>("api")
                    .expect("Same reason");
                trinitrix_api
                    .set("register_function", wrapped_register_function)
                    .expect("This should work");
            }

            Mutex::new(lua)
        })
        .lock()
        .await;
    lua
}

async fn exec_lua(lua_code: &str, event_call_tx: mpsc::Sender<Event>) -> Result<()> {
    let lua = initialize_lua(event_call_tx.clone()).await;

    info!("Recieved code to execute: `{}`, executing...", &lua_code);
    let output = lua.load(lua_code).eval_async::<Value>().await;
    match output {
        Ok(out) => {
            let to_string_fn: mlua::Function =
                lua.globals().get("tostring").expect("This always exists");
            let output: String = to_string_fn.call(out).expect("tostring should not error");
            info!("Lua code `{}` evaluated to: `{}`", lua_code, &output);

            if output != "nil" {
                event_call_tx
                    .send(Event::CommandEvent(
                        Command::Trinitrix(Api(Raw(DisplayOutput(output)))),
                        None,
                    ))
                    .await
                    .context("Failed to send lua output command")?
            }
        }
        Err(err) => {
            error!("Lua code `{}` returned error: `{}`", lua_code, err);
            event_call_tx
                .send(Event::CommandEvent(
                    Command::Trinitrix(Api(Raw(RaiseError(err.to_string())))),
                    None,
                ))
                .await?;
        }
    };
    Ok(())
}
