use std::str::FromStr;

use crate::{
    app::{
        command_interface::{
            trinitrix::{
                api::{keymaps::Keymaps, raw::Raw, ui::Ui, Api},
                Trinitrix,
            },
            Commands,
        },
        events::EventStatus,
        status::State,
        App,
    },
    trinitrix::api::ui::Mode,
    ui::central::InputPosition,
};
use anyhow::Result;
use cli_log::{info, trace, warn};
use crossterm::event::Event;
use keymaps::{
    key_repr::{Key, Keys},
    trie::Node,
};
use trixy::oneshot;

pub async fn handle(
    app: &mut App<'_>,
    command: &Commands,

    // FIXME(@soispha): The `String` is temporary <2024-05-03>
    output_callback: Option<oneshot::Sender<String>>,
) -> Result<EventStatus> {
    // A command can both return _status output_ (what you would normally print to stderr)
    // and _main output_ (the output which is normally printed to stdout).
    // We simulate these by returning the main output to the Lua function, and printing the
    // status output to a status UI field.
    //
    // Every function should return some status output to show the user, that something is
    // happening, while only some functions return some value to the main output, as this
    // is reserved for functions called only for their output (for example `greet()`).
    macro_rules! send_status_output {
        ($str:expr) => {
             app.status.add_status_message($str.to_owned())
        };
        ($str:expr, $($args:ident),+) => {
             app.status.add_status_message(format!($str, $($args),+))
        };
    }
    macro_rules! send_error_output {
        ($str:expr) => {
             app.status.add_error_message($str.to_owned())
        };
        ($str:expr, $($args:ident),+) => {
             app.status.add_error_message(format!($str, $($args),+))
        };
    }
    // macro_rules! send_main_output {
    //     ($str:expr) => {
    //         if let Some(sender) = output_callback {
    //             sender
    //                 .send(CommandTransferValue::from($str))
    //                 .map_err(|e| Error::msg(format!("Failed to send command main output: `{}`", e)))?;
    //         }
    //     };
    //     ($str:expr, $($args:ident),+) => {
    //         if let Some(sender) = output_callback {
    //             sender
    //                 .send(CommandTransferValue::from(format!($str, $($args),+)))
    //                 .map_err(|e| Error::msg(format!("Failed to send command main output: `{}`", e)))?;
    //         }
    //     };
    // }

    trace!("Handling command: {:#?}", command);

    Ok(match command {
        Commands::Trinitrix(trinitrix) => match trinitrix {
            Trinitrix::Stdi(_) => {
                // No-op I guess
                EventStatus::Ok
            }
            Trinitrix::Api(api) => match api {
                Api::exit => {
                    send_status_output!("Terminating the application..");
                    warn!("Terminating the application");
                    EventStatus::Terminate
                }
                Api::room_message_send { message } => {
                    if let Some(room) = app.status.room_mut() {
                        room.send(message.to_string()).await?;
                        send_status_output!("Sent message: `{}`", message);
                    } else {
                        // FIXME(@soispha): This should raise an error within lua, as it would
                        // otherwise be very confusing <2023-09-20>
                        warn!(
                            "Can't send message: `{}`, as there is no open room!",
                            &message
                        );
                    }

                    EventStatus::Ok
                }
                Api::Ui(ui) => match ui {
                    Ui::set_mode { mode } => match mode {
                        Mode::Normal => {
                            app.status.set_state(State::Normal);
                            send_status_output!("Set input mode to Normal");
                            EventStatus::Ok
                        }
                        Mode::Insert => {
                            app.status.set_state(State::Insert);
                            app.ui.set_input_position(InputPosition::MessageCompose);
                            send_status_output!("Set input mode to Insert");
                            EventStatus::Ok
                        }
                        Mode::Command => {
                            app.ui.cli_enable();
                            app.status.set_state(State::Command);
                            send_status_output!("Set input mode to CLI");
                            EventStatus::Ok
                        }
                    },
                    Ui::cycle_planes => {
                        app.ui.cycle_main_input_position();
                        send_status_output!("Switched main input position");
                        EventStatus::Ok
                    }
                    Ui::cycle_planes_rev => {
                        app.ui.cycle_main_input_position_rev();
                        send_status_output!("Switched main input position; reversed");
                        EventStatus::Ok
                    }
                },
                Api::Keymaps(keymaps) => match keymaps {
                    Keymaps::add {
                        mode,
                        key,
                        callback,
                    } => {
                        info!("Will add a keymapping for '{}' in mode '{}'", key, mode);
                        mode.as_str().chars().for_each(|char| {
                            info!("Setting keymaping ('{}') for mode '{}'", key, char);
                            let parsed_keys = key
                                .as_str()
                                .parse::<Keys>()
                                .map_err(|err| {
                                    send_error_output!(err.to_string());
                                })
                                .expect("We dealt with the error");

                            match State::from_char(&char) {
                                Ok(state) => {
                                    info!("Set for state '{}'", state);
                                    let trie;
                                    if let Some(collected_trie) = app.key_mappings.get_mut(&state) {
                                        trie = collected_trie;
                                    } else {
                                        app.key_mappings.insert(state.clone(), Node::new());
                                        trie = app
                                            .key_mappings
                                            .get_mut(&state)
                                            .expect("Should be set");
                                    }
                                    trie.insert(&parsed_keys, callback.to_owned())
                                        .map_err(|err| {
                                            send_error_output!(format!("{:#?}", err));
                                        })
                                        .expect("We already dealt with the error")
                                }
                                Err(err) => send_error_output!(err.to_string()),
                            };
                        });

                        EventStatus::Ok
                    }
                    // FIXME(@soispha): It would be nice to have these functions, but well..
                    // someone needs to write them <2024-05-03>
                    Keymaps::remove { mode, key } => todo!(),
                    Keymaps::get { mode } => todo!(),
                },
                Api::Raw(raw) => match raw {
                    Raw::raise_error { error_message } => {
                        send_error_output!(error_message.to_string());
                        EventStatus::Ok
                    }
                    Raw::display_output { output_message } => {
                        // TODO(@Soispha): This is only used to show the Lua command output to the user.
                        // Lua commands already receive the output. This should probably be communicated
                        // better, should it?
                        send_status_output!(output_message.to_string());
                        EventStatus::Ok
                    }
                    Raw::send_input_unprocessed { input } => {
                        let output = match app.status.state() {
                            State::Insert => {
                                let key = Key::from_str(input.as_str())?;
                                let cross_input: Event = key.try_into()?;
                                app.ui
                                    .message_compose
                                    .input(tui_textarea::Input::from(cross_input));
                                EventStatus::Ok
                            }
                            State::Command => {
                                let key = Key::from_str(input.as_str())?;
                                let cross_input: Event = key.try_into()?;
                                app.ui
                                    .cli
                                    .as_mut()
                                    .expect("This should exist, when the state is 'Command'")
                                    .input(tui_textarea::Input::from(cross_input));
                                EventStatus::Ok
                            }
                            State::Normal
                            | State::Setup
                            | State::KeyInputPending {
                                old_state: _,
                                pending_keys: _,
                            } => EventStatus::Ok,
                        };
                        output
                    }
                    Raw::Private(private) => {
                        // no-op, this was used to store functions (not so sure, if we need it
                        // any longer)
                        EventStatus::Ok
                    }
                },
            },
        },
        // Command::Print(output) => {
        //     let output_str: String = output.to_string();
        //     send_status_output!(output_str);
        //     EventStatus::Ok
        // }
        //
        // Command::Trinitrix(trinitrix) => match trinitrix {
        //     Trinitrix::Debug(debug) => match debug {
        //         Debug::Greet(msg) => {
        //             send_main_output!("Greeting, {}!", msg);
        //             EventStatus::Ok
        //         }
        //         Debug::GreetMultiple => {
        //             let mut table: Table = HashMap::new();
        //             table.insert("UserId".to_owned(), CommandTransferValue::Integer(2));
        //             table.insert(
        //                 "UserName".to_owned(),
        //                 CommandTransferValue::String("James".to_owned()),
        //             );
        //
        //             let mut second_table: Table = HashMap::new();
        //             second_table.insert("interface".to_owned(), CommandTransferValue::Integer(3));
        //             second_table.insert("api".to_owned(), CommandTransferValue::Boolean(true));
        //             table.insert(
        //                 "Versions".to_owned(),
        //                 CommandTransferValue::Table(second_table),
        //             );
        //             send_main_output!(table);
        //             EventStatus::Ok
        //         }
        //     },
        //     Trinitrix::Api(api) => match api {
        //         Api::RoomMessageSend(msg) => {
        //         }
        //         Api::Help(_) => todo!(),
        //         Api::Ui(ui) => match ui {
        //             Ui::CommandLineShow => {
        //             }
        //             Ui::CommandLineHide => {
        //                 app.ui.cli_disable();
        //                 send_status_output!("CLI offline");
        //                 EventStatus::Ok
        //             }
        //             Ui::CyclePlanes => {
        //             }
        //             Ui::CyclePlanesRev => {
        //             }
        //             Ui::SetModeNormal => {
        //             }
        //             Ui::SetModeInsert => {
        //             }
        //         },
        //         Api::Keymaps(keymaps) => match keymaps {
        //             Keymaps::Add((mode, key, callback)) => {
        //             }
        //             // TODO(@soispha): Well.., we should probably add these functions: <2023-10-15>
        //             Keymaps::Remove((mode, key)) => todo!(),
        //             Keymaps::Get(mode) => todo!(),
        //         },
        //         Api::Raw(raw) => match raw {
        //             Raw::RaiseError(err) => {
        //             }
        //             Raw::DisplayOutput(output) => {
        //             }
        //             Raw::Private(_) => {
        //             }
        //             Raw::SendInputUnprocessed(char) =>
        //         },
        //     },
        //     Trinitrix::Std(_) => {
        //         // no-op, read the comment about it in the `command_list`
        //         EventStatus::Ok
        //     }
        // },
    })
}
