use anyhow::Result;
use cli_log::info;
use crossterm::event::Event as CrosstermEvent;
use keymaps::key_repr::{Key, Keys};

use crate::app::{
    command_interface::{
        trinitrix::{
            api::{raw::Raw, Api},
            Trinitrix,
        },
        Commands,
    },
    events::{Event, EventStatus},
    status::State,
    App,
};

pub async fn handle(app: &mut App<'_>, input_event: &CrosstermEvent) -> Result<EventStatus> {
    async fn default(converted_key: Key, app: &mut App<'_>, old_state: &State) -> Result<()> {
        info!(
            "No keymaps exist for key ('{}'), passing it along..",
            converted_key
        );
        if let State::KeyInputPending {
            old_state: _,
            pending_keys,
        } = app.status.state().clone()
        {
            for key in pending_keys {
                app.tx
                    .send(Event::CommandEvent(
                        Commands::Trinitrix(Trinitrix::Api(Api::Raw(
                            Raw::send_input_unprocessed {
                                input: key.to_string_repr().into(),
                            },
                        ))),
                        None,
                    ))
                    .await?;
            }
            app.status.set_state(old_state.to_owned());
        }
        // Just let the input event slip through if no keymap matches
        app.tx
            .send(Event::CommandEvent(
                Commands::Trinitrix(Trinitrix::Api(Api::Raw(Raw::send_input_unprocessed {
                    input: converted_key.to_string_repr().into(),
                }))),
                None,
            ))
            .await?;
        Ok(())
    }

    if let CrosstermEvent::Key(_) = input_event {
        //    r
        //    |
        //    a
        //   / \
        //  b   a
        //      |
        //      c
        //
        //
        //
        //   r->a->a: Some([a]) -> a.is_child() && a.is_terminal() ? call a : *key input pending*
        //   r->a->b: Some([b]) -> b.is_child() && b.is_terminal() ? *call b* : key input pending
        //   r->a->c: None      -> continue
        //   r->a->a->c: Some([c]) -> c.is_child() && c.is_terminal() ? *call c* : key input pending
        //
        //   r->a: Some([a, b]) -> key input pending

        let converted_key: Key = input_event.try_into()?;
        info!("Received input to handle: '{}'", converted_key);
        let mut converted_keys: Keys = Keys::new(converted_key);

        let mut old_state = app.status.state().clone();
        if let State::KeyInputPending {
            old_state: old,
            pending_keys,
        } = app.status.state().clone()
        {
            info!("Found KeyInputPending mode!");
            old_state = *old;
            converted_keys = pending_keys.join(converted_key);
        }

        if let Some(key_maps) = app.key_mappings.get(&old_state) {
            if let Some((possible_key_maps, should_call)) = key_maps.get(&converted_keys) {
                info!("possible key maps: {:#?}", possible_key_maps);

                if possible_key_maps.len() == 1 {
                    let possible_key_map = possible_key_maps.get(0).expect("The len is 1");

                    if possible_key_map.is_child() && possible_key_map.is_terminal() && should_call
                    {
                        let function = possible_key_map
                            .value()
                            .expect("This node is terminal and a child, it should have a value");

                        function();
                        // app.tx.send(Event::Function(*function)).await?;
                        app.status.set_state(old_state.to_owned());
                    } else {
                        // The choice does not have a value attached to it (might be a waypoint)
                        app.status.set_state(State::KeyInputPending {
                            old_state: Box::new(old_state.to_owned()),
                            pending_keys: converted_keys,
                        });
                    }
                } else {
                    app.status.set_state(State::KeyInputPending {
                        old_state: Box::new(old_state.to_owned()),
                        pending_keys: converted_keys,
                    })
                }
            } else {
                default(converted_key, app, &old_state).await?
            }
        } else {
            default(converted_key, app, &old_state).await?
        }
    }
    Ok(EventStatus::Ok)
}
