use anyhow::{Context, Result};
use crossterm::event::{Event as CrosstermEvent, KeyCode, KeyEvent, KeyModifiers};

use crate::{
    app::{
        command_interface::{
            Api::{Exit, Ui, RoomMessageSend},
            Command,
            Trinitrix::Api,
            Ui::{CommandLineShow, CyclePlanes, CyclePlanesRev, SetModeInsert, SetModeNormal},
        },
        events::event_types::{Event, EventStatus},
        App,
    },
    ui::central,
};

pub async fn handle_command(
    app: &mut App<'_>,
    input_event: &CrosstermEvent,
) -> Result<EventStatus> {
    if let Some(cli) = &app.ui.cli {
        match input_event {
            CrosstermEvent::Key(KeyEvent {
                code: KeyCode::Esc, ..
            }) => {
                app.tx
                    .send(Event::CommandEvent(
                        Command::Trinitrix(Api(Ui(SetModeNormal))),
                        None,
                    ))
                    .await?;
            }
            CrosstermEvent::Key(KeyEvent {
                code: KeyCode::Enter,
                ..
            }) => {
                let ci_event = cli
                    .lines()
                    .get(0)
                    .expect(
                        "One line always exists,
                     and others can't exists
                     because we collect on
                     enter",
                    )
                    .to_owned();
                app.tx
                    .send(Event::LuaCommand(ci_event))
                    .await
                    .context("Failed to send lua command to internal event stream")?;
            }
            _ => {
                app.ui
                    .cli
                    .as_mut()
                    .expect("This is already checked")
                    .input(tui_textarea::Input::from(input_event.to_owned()));
            }
        }
    } else {
        unreachable!("The cli should not be active while no cli is defined");
    }
    Ok(EventStatus::Ok)
}
pub async fn handle_normal(app: &mut App<'_>, input_event: &CrosstermEvent) -> Result<EventStatus> {
    match input_event {
        CrosstermEvent::Key(KeyEvent {
            code: KeyCode::Char('q'),
            ..
        }) => {
            app.tx
                .send(Event::CommandEvent(Command::Trinitrix(Api(Exit)), None))
                .await?;
        }
        CrosstermEvent::Key(KeyEvent {
            code: KeyCode::Tab, ..
        }) => {
            app.tx
                .send(Event::CommandEvent(
                    Command::Trinitrix(Api(Ui(CyclePlanes))),
                    None,
                ))
                .await?;
        }
        CrosstermEvent::Key(KeyEvent {
            code: KeyCode::BackTab,
            ..
        }) => {
            app.tx
                .send(Event::CommandEvent(
                    Command::Trinitrix(Api(Ui(CyclePlanesRev))),
                    None,
                ))
                .await?;
        }
        CrosstermEvent::Key(KeyEvent {
            code: KeyCode::Char(':'),
            ..
        }) => {
            app.tx
                .send(Event::CommandEvent(
                    Command::Trinitrix(Api(Ui(CommandLineShow))),
                    None,
                ))
                .await?;
        }
        CrosstermEvent::Key(KeyEvent {
            code: KeyCode::Char('i'),
            ..
        }) => {
            app.tx
                .send(Event::CommandEvent(
                    Command::Trinitrix(Api(Ui(SetModeInsert))),
                    None,
                ))
                .await?;
        }
        input => match app.ui.input_position() {
            central::InputPosition::Rooms => {
                match input {
                    CrosstermEvent::Key(KeyEvent {
                        code: KeyCode::Up, ..
                    }) => {
                        let i = match app.ui.rooms_state.selected() {
                            Some(cur) => {
                                if cur > 0 {
                                    cur - 1
                                } else {
                                    cur
                                }
                            }
                            None => 0,
                        };
                        app.ui.rooms_state.select(Some(i));
                        app.status.set_room_by_index(i)?;
                    }
                    CrosstermEvent::Key(KeyEvent {
                        code: KeyCode::Down,
                        ..
                    }) => {
                        let i = match app.ui.rooms_state.selected() {
                            Some(cur) => {
                                if cur < app.status.rooms().len() - 1 {
                                    cur + 1
                                } else {
                                    cur
                                }
                            }
                            None => 0,
                        };
                        app.ui.rooms_state.select(Some(i));
                        app.status.set_room_by_index(i)?;
                    }
                    _ => (),
                };
            }
            central::InputPosition::Messages => {
                match input {
                    CrosstermEvent::Key(KeyEvent {
                        code: KeyCode::Up, ..
                    }) => {
                        match app.status.room_mut() {
                            Some(room) => {
                                let len = room.timeline().len();
                                let i = match room.view_scroll() {
                                    Some(i) => i + 1,
                                    None => 0,
                                };
                                if i < len {
                                    room.set_view_scroll(Some(i))
                                }
                                if i <= len - 5 {
                                    room.poll_old_timeline().await?;
                                }
                            }
                            None => (),
                        };
                    }
                    CrosstermEvent::Key(KeyEvent {
                        code: KeyCode::Down,
                        ..
                    }) => {
                        match app.status.room_mut() {
                            Some(room) => {
                                match room.view_scroll() {
                                    Some(i) => {
                                        if i == 0 {
                                            room.set_view_scroll(None);
                                        } else {
                                            room.set_view_scroll(Some(i - 1));
                                        }
                                    }
                                    None => (),
                                };
                            }
                            None => (),
                        };
                    }
                    _ => (),
                };
            }

            _ => (),
        },
    };
    Ok(EventStatus::Ok)
}

pub async fn handle_insert(app: &mut App<'_>, event: &CrosstermEvent) -> Result<EventStatus> {
    match event {
        CrosstermEvent::Key(KeyEvent {
            code: KeyCode::Esc, ..
        }) => {
            app.tx
                .send(Event::CommandEvent(
                    Command::Trinitrix(Api(Ui(SetModeNormal))),
                    None,
                ))
                .await?;
        }
        CrosstermEvent::Key(KeyEvent {
            code: KeyCode::Enter,
            modifiers: KeyModifiers::ALT,
            ..
        }) => {
            app.tx
                .send(Event::CommandEvent(
                    Command::Trinitrix(Api(RoomMessageSend(
                        app.ui.message_compose.lines().join("\n"),
                    ))),
                    None,
                ))
                .await?;
            app.ui.message_compose_clear();
        }
        _ => {
            app.ui
                .message_compose
                .input(tui_textarea::Input::from(event.to_owned()));
        }
    };

    Ok(EventStatus::Ok)
}
