use anyhow::{bail, Context, Result};
use crossterm::event::{Event as CrosstermEvent, KeyCode, KeyEvent};

use crate::{
    app::{events::EventStatus, App},
    ui::setup,
};

pub async fn handle(app: &mut App<'_>, input_event: &CrosstermEvent) -> Result<EventStatus> {
    let ui = match &mut app.ui.setup_ui {
        Some(ui) => ui,
        None => bail!("SetupUI instance not found"),
    };

    match input_event {
        CrosstermEvent::Key(KeyEvent {
            code: KeyCode::Esc, ..
        }) => return Ok(EventStatus::Terminate),
        CrosstermEvent::Key(KeyEvent {
            code: KeyCode::Tab, ..
        }) => {
            ui.cycle_input_position();
        }
        CrosstermEvent::Key(KeyEvent {
            code: KeyCode::BackTab,
            ..
        }) => {
            ui.cycle_input_position_rev();
        }
        CrosstermEvent::Key(KeyEvent {
            code: KeyCode::Enter,
            ..
        }) => {
            match ui.input_position() {
                setup::InputPosition::Ok => {
                    let homeserver = ui.homeserver.lines()[0].clone();
                    let username = ui.username.lines()[0].clone();
                    let password = ui.password_data.lines()[0].clone();
                    app.login(&homeserver, &username, &password)
                        .await
                        .context("Failed to login")?;
                    // We bailed in the line above, thus login must have succeeded
                    return Ok(EventStatus::Finished);
                }
                _ => ui.cycle_input_position(),
            };
        }
        input => match ui.input_position() {
            setup::InputPosition::Homeserver => {
                ui.homeserver.input(input.to_owned());
            }
            setup::InputPosition::Username => {
                ui.username.input(input.to_owned());
            }
            setup::InputPosition::Password => {
                let textarea_input = tui_textarea::Input::from(input.to_owned());
                ui.password_data.input(textarea_input.clone());
                match textarea_input.key {
                    tui_textarea::Key::Char(_) => {
                        ui.password.input(tui_textarea::Input {
                            key: tui_textarea::Key::Char('*'),
                            ctrl: false,
                            alt: false,
                        });
                    }
                    _ => {
                        ui.password.input(textarea_input);
                    }
                }
            }
            _ => (),
        },
    };
    Ok(EventStatus::Ok)
}
