mod handlers;
pub mod listeners;

use anyhow::{Context, Result};

use crate::app::{command_interface::Commands, status::State, App};
use cli_log::{trace, warn};
use crossterm::event::Event as CrosstermEvent;
use handlers::{command, input, lua_command, matrix, setup};

#[derive(Debug)]
pub enum Event {
    InputEvent(CrosstermEvent),
    MatrixEvent(matrix_sdk::deserialized_responses::SyncResponse),
    // FIXME(@soispha): The `String` is also wrong <2024-05-03>
    CommandEvent(Commands, Option<trixy::oneshot::Sender<String>>),
    LuaCommand(String),
}

impl Event {
    pub async fn handle(self, app: &mut App<'_>) -> Result<EventStatus> {
        trace!("Received event to handle: `{:#?}`", &self);
        match self {
            Event::MatrixEvent(event) => matrix::handle(app, &event)
                .await
                .with_context(|| format!("Failed to handle matrix event: `{:#?}`", event)),

            Event::CommandEvent(event, callback_tx) => command::handle(app, &event, callback_tx)
                .await
                .with_context(|| format!("Failed to handle command event: `{:#?}`", event)),

            Event::LuaCommand(lua_code) => {
                warn!(
                    "Got lua code to execute, but no exectuter is available:\n{}",
                    lua_code
                );
                Ok(EventStatus::Ok)
            }
            // lua_command::handle(app, lua_code.to_owned())
            // .await
            // .with_context(|| format!("Failed to handle lua code: `{}`", lua_code)),
            // Event::Function(function) => function::handle(app, function.to_owned())
            //     .await
            //     .with_context(|| format!("Failed to handle function: `{}`", function)),
            Event::InputEvent(event) => match app.status.state() {
                State::Setup => setup::handle(app, &event).await.with_context(|| {
                    format!("Failed to handle input (setup) event: `{:#?}`", event)
                }),
                _ => input::handle(app, &event).await.with_context(|| {
                    format!("Failed to handle input (non-setup) event: `{:#?}`", event)
                }),
            },
        }
    }
}

#[derive(Debug)]
pub enum EventStatus {
    Ok,
    Finished,
    Terminate,
}
