pub mod command_interface;
pub mod config;
pub mod events;
pub mod status;

use std::{
    collections::HashMap,
    ffi::c_int,
    path::{Path, PathBuf},
    sync::OnceLock,
};

use anyhow::{Context, Error, Result};
use cli_log::{info, warn};
use crossterm::{
    event::DisableMouseCapture,
    execute,
    terminal::{disable_raw_mode, LeaveAlternateScreen},
};
use directories::ProjectDirs;
use keymaps::trie::Node;
use libloading::{Library, Symbol};
use matrix_sdk::Client;
use tokio::sync::mpsc::{self, Sender};
use tokio_util::sync::CancellationToken;

// use self::command_interface::{
// lua_command_manager::LuaCommandManager,
// };

use crate::{
    accounts::{Account, AccountsManager},
    app::{
        events::{Event, EventStatus},
        status::{State, Status},
    },
    ui::{central, setup},
};

pub struct App<'runtime> {
    ui: central::UI<'runtime>,
    accounts_manager: AccountsManager,
    status: Status,

    tx: mpsc::Sender<Event>,
    rx: mpsc::Receiver<Event>,

    input_listener_killer: CancellationToken,
    matrix_listener_killer: CancellationToken,

    // lua: LuaCommandManager,
    project_dirs: ProjectDirs,

    key_mappings: HashMap<State, Node<extern "C" fn()>>,
}

pub static COMMAND_TRANSMITTER: OnceLock<Sender<Event>> = OnceLock::new();

impl App<'_> {
    pub fn new() -> Result<Self> {
        let path: &std::path::Path = Path::new("userdata/accounts.json");
        let config = if path.exists() {
            info!("Reading account config (userdata/accounts.json)");
            Some(std::fs::read_to_string(path)?)
        } else {
            None
        };

        let (tx, rx) = mpsc::channel(256);

        COMMAND_TRANSMITTER
            .set(tx.clone())
            .expect("The cell should always be empty at this point");

        Ok(Self {
            ui: central::UI::new()?,
            accounts_manager: AccountsManager::new(config)?,
            status: Status::new(None),

            tx: tx.clone(),
            rx,
            input_listener_killer: CancellationToken::new(),
            matrix_listener_killer: CancellationToken::new(),

            // lua: LuaCommandManager::new(tx),

            // TODO: We probably want to populate the strings below a bit more <2023-09-09>
            project_dirs: ProjectDirs::from("", "", "trinitrix").context(
                "Failed to allocate project direcectory paths, \
                          please ensure your $HOME is set correctly",
            )?,
            key_mappings: HashMap::new(),
        })
    }

    pub async fn run(
        &mut self,
        cli_lua_config_file: Option<PathBuf>,
        plugin_path: Option<PathBuf>,
    ) -> Result<()> {
        // Spawn input event listener
        tokio::task::spawn(events::listeners::input::poll(
            self.tx.clone(),
            self.input_listener_killer.clone(),
        ));

        let config_dir = self.project_dirs.config_dir();
        let mut lua_config_file = None;
        if config_dir
            .try_exists()
            .context("Failed to search for a config file")?
        {
            lua_config_file = Some(config_dir.join("init.lua"));
            info!("Found config dir: `{}`", config_dir.display());
        } else {
            warn!("No config dir found!");
        }

        if let Some(config_file) = cli_lua_config_file {
            config::lua::load(self.tx.clone(), config_file).await?;
        } else if let Some(config_file) = lua_config_file {
            info!("Searching for config file at `{}`", config_file.display());
            if config_file.try_exists().with_context(|| {
                format!(
                    "Failed to check, if the lua config file at `{}` exists",
                    config_file.display()
                )
            })? {
                config::lua::load(self.tx.clone(), config_file).await?;
            } else {
                warn!("No config file found!");
            }
        }

        if let Some(plugin) = plugin_path {
            info!("Loading plugin_main() from {}", plugin.display());

            unsafe {
                let lib = Library::new(plugin).context("Failed to load plugin")?;
                let func: Symbol<unsafe fn() -> c_int> = lib
                    .get(b"plugin_main")
                    .context("Plugin does not have a 'plugin_main' symbol")?;

                info!("Starting plugin");
                let out = func();
                info!("Plugin finished with: {}", out);
            }
        }

        if self.account().is_err() {
            info!("No saved sessions found -> jumping into setup");
            self.setup().await?;
        } else {
            self.accounts_manager.restore().await?;
            self.init_account().await?;
        }

        self.status.set_state(State::Normal);

        loop {
            self.ui.update(&self.status).await?;

            let event = self.rx.recv().await.context("Failed to get next event")?;
            match event.handle(self).await? {
                EventStatus::Ok => (),
                EventStatus::Terminate => break,
                _ => todo!(),
            };
        }

        self.input_listener_killer.cancel();
        Ok(())
    }

    async fn setup(&mut self) -> Result<()> {
        self.ui.setup_ui = Some(setup::UI::new());

        self.status.set_state(State::Setup);

        loop {
            self.ui.update_setup().await?;

            let event = self.rx.recv().await.context("Failed to get next event")?;

            match event.handle(self).await? {
                EventStatus::Ok => (),
                EventStatus::Finished => return Ok(()),
                EventStatus::Terminate => return Err(Error::msg("Terminated by user")),
            }
        }
    }

    pub async fn init_account(&mut self) -> Result<()> {
        let client = match self.client() {
            Some(c) => c,
            None => return Err(Error::msg("failed to get current client")),
        }
        .clone();

        self.matrix_listener_killer.cancel();
        self.matrix_listener_killer = CancellationToken::new();

        // Spawn Matrix Event Listener
        tokio::task::spawn(events::listeners::matrix::poll(
            self.tx.clone(),
            self.matrix_listener_killer.clone(),
            client.clone(),
        ));

        // Reset Status
        self.status = Status::new(Some(client));

        let account = self.account()?;
        let name = account.name().clone();
        let user_id = account.user_id().clone();
        self.status.set_account_name(name);
        self.status.set_account_user_id(user_id);

        for (_, room) in self.status.rooms_mut() {
            room.update_name().await?;
            for _ in 0..3 {
                room.poll_old_timeline().await?;
            }
        }

        info!("Initializing client for the current account");

        Ok(())
    }

    pub async fn switch_account(&mut self, account_id: u32) -> Result<()> {
        Ok(())
    }

    pub async fn login(
        &mut self,
        homeserver: &String,
        username: &String,
        password: &String,
    ) -> Result<()> {
        self.accounts_manager
            .add(homeserver, username, password)
            .await?;
        self.init_account().await?;
        Ok(())
    }

    pub fn account(&self) -> Result<&Account> {
        let account = self.accounts_manager.current();
        match account {
            None => Err(Error::msg("failed to resolve current account")),
            Some(a) => Ok(a),
        }
    }

    pub fn client(&self) -> Option<&Client> {
        self.accounts_manager.client()
    }
}
