pub mod update;

use std::io::Stdout;

use anyhow::{bail, Context, Result};
use cli_log::{info, warn};
use crossterm::{
    event::DisableMouseCapture,
    execute,
    terminal::{disable_raw_mode, LeaveAlternateScreen},
};
use tui::{
    backend::CrosstermBackend,
    style::Color,
    widgets::{Block, Borders, ListState},
    Terminal,
};
use tui_textarea::TextArea;
pub use update::*;

use super::setup;
use crate::ui::{terminal_prepare, textarea_activate, textarea_inactivate};

#[derive(Clone, Copy, PartialEq)]
pub enum InputPosition {
    Status,
    Rooms,
    Messages,
    MessageCompose,
    CommandMonitor,
    RoomInfo,
    CLI,
}

impl InputPosition {
    // calculate to widgets colors, based of which widget is currently selected
    pub fn colors(
        &self,
        mut cli: &mut Option<TextArea<'_>>,
        mut message_compose: &mut TextArea<'_>,
    ) -> Vec<Color> {
        match self {
            InputPosition::Status => {
                textarea_inactivate(&mut message_compose);
                if let Some(cli) = &mut cli {
                    textarea_inactivate(cli);
                }
                vec![
                    Color::White,
                    Color::DarkGray,
                    Color::DarkGray,
                    Color::DarkGray,
                    Color::DarkGray,
                    Color::DarkGray,
                ]
            }
            InputPosition::Rooms => {
                textarea_inactivate(&mut message_compose);
                if let Some(cli) = &mut cli {
                    textarea_inactivate(cli);
                }
                vec![
                    Color::DarkGray,
                    Color::White,
                    Color::DarkGray,
                    Color::DarkGray,
                    Color::DarkGray,
                    Color::DarkGray,
                ]
            }
            InputPosition::Messages => {
                textarea_inactivate(&mut message_compose);
                if let Some(cli) = &mut cli {
                    textarea_inactivate(cli);
                }
                vec![
                    Color::DarkGray,
                    Color::DarkGray,
                    Color::White,
                    Color::DarkGray,
                    Color::DarkGray,
                    Color::DarkGray,
                ]
            }
            InputPosition::MessageCompose => {
                textarea_activate(&mut message_compose);
                if let Some(cli) = &mut cli {
                    textarea_inactivate(cli);
                }
                vec![
                    Color::DarkGray,
                    Color::DarkGray,
                    Color::DarkGray,
                    Color::DarkGray,
                    Color::DarkGray,
                    Color::DarkGray,
                ]
            }
            InputPosition::RoomInfo => {
                textarea_inactivate(&mut message_compose);
                if let Some(cli) = &mut cli {
                    textarea_inactivate(cli);
                }
                vec![
                    Color::DarkGray,
                    Color::DarkGray,
                    Color::DarkGray,
                    Color::DarkGray,
                    Color::White,
                    Color::DarkGray,
                ]
            }
            InputPosition::CLI => {
                textarea_inactivate(&mut message_compose);
                if let Some(cli) = &mut cli {
                    textarea_activate(cli);
                }
                vec![
                    Color::DarkGray,
                    Color::DarkGray,
                    Color::DarkGray,
                    Color::DarkGray,
                    Color::DarkGray,
                    Color::DarkGray,
                ]
            }
            InputPosition::CommandMonitor => {
                textarea_inactivate(&mut message_compose);
                if let Some(cli) = &mut cli {
                    textarea_inactivate(cli);
                }
                vec![
                    Color::DarkGray,
                    Color::DarkGray,
                    Color::DarkGray,
                    Color::DarkGray,
                    Color::DarkGray,
                    Color::White,
                ]
            }
        }
    }
}

pub struct UI<'a> {
    terminal: Terminal<CrosstermBackend<Stdout>>,
    input_position: InputPosition,
    pub rooms_state: ListState,
    pub message_compose: TextArea<'a>,
    pub cli: Option<TextArea<'a>>,

    pub setup_ui: Option<setup::UI<'a>>,
}

impl Drop for UI<'_> {
    fn drop(&mut self) {
        info!("Destructing UI");
        disable_raw_mode().expect("While destructing UI -> Failed to disable raw mode");
        execute!(
            self.terminal.backend_mut(),
            LeaveAlternateScreen,
            DisableMouseCapture
        ).expect("While destructing UI -> Failed execute backend commands (LeaveAlternateScreen and DisableMouseCapture)");
        self.terminal
            .show_cursor()
            .expect("While destructing UI -> Failed to re-enable cursor");
    }
}

impl UI<'_> {
    pub fn new() -> Result<Self> {
        let stdout = terminal_prepare().context("Falied to prepare terminal")?;
        let backend = CrosstermBackend::new(stdout);
        let mut terminal = Terminal::new(backend)?;

        terminal.clear()?;

        let mut message_compose = TextArea::default();
        message_compose.set_block(
            Block::default()
                .title("Message Compose (send: <Alt>+<Enter>)")
                .borders(Borders::ALL),
        );

        info!("Initialized UI");

        Ok(Self {
            terminal,
            input_position: InputPosition::Rooms,
            rooms_state: ListState::default(),
            message_compose,
            cli: None,
            setup_ui: None,
        })
    }

    pub fn cycle_main_input_position(&mut self) {
        self.input_position = match self.input_position {
            InputPosition::Status => InputPosition::Rooms,
            InputPosition::Rooms => InputPosition::Messages,
            InputPosition::Messages => InputPosition::MessageCompose,
            InputPosition::MessageCompose => InputPosition::CommandMonitor,
            InputPosition::CommandMonitor => InputPosition::RoomInfo,
            InputPosition::RoomInfo => match self.cli {
                Some(_) => InputPosition::CLI,
                None => InputPosition::Status,
            },
            InputPosition::CLI => InputPosition::Status,
        };
    }

    pub fn cycle_main_input_position_rev(&mut self) {
        self.input_position = match self.input_position {
            InputPosition::Status => match self.cli {
                Some(_) => InputPosition::CLI,
                None => InputPosition::RoomInfo,
            },
            InputPosition::Rooms => InputPosition::Status,
            InputPosition::Messages => InputPosition::Rooms,
            InputPosition::MessageCompose => InputPosition::Messages,
            InputPosition::RoomInfo => InputPosition::CommandMonitor,
            InputPosition::CommandMonitor => InputPosition::MessageCompose,
            InputPosition::CLI => InputPosition::RoomInfo,
        };
    }

    pub fn set_input_position(&mut self, position: InputPosition) {
        self.input_position = position;
    }

    pub fn input_position(&self) -> &InputPosition {
        &self.input_position
    }

    pub fn message_compose_clear(&mut self) {
        self.message_compose = TextArea::default();
        self.message_compose.set_block(
            Block::default()
                .title("Message Compose (send: <Alt>+<Enter>)")
                .borders(Borders::ALL),
        );
    }

    pub fn set_command_output(&mut self, output: &str) {
        info!("Setting output to: `{}`", output);
        if let Some(_) = self.cli {
            let cli = Some(TextArea::from([output]));
            self.cli = cli;
        } else {
            warn!("Failed to set output");
        }
    }

    pub fn cli_enable(&mut self) {
        self.input_position = InputPosition::CLI;
        if self.cli.is_some() {
            return;
        }
        let mut cli = TextArea::default();
        cli.set_block(Block::default().borders(Borders::ALL));
        self.cli = Some(cli);
    }

    pub fn cli_disable(&mut self) {
        if self.input_position == InputPosition::CLI {
            self.cycle_main_input_position();
        }
        self.cli = None;
    }

    pub async fn update_setup(&mut self) -> Result<()> {
        let setup_ui = match &mut self.setup_ui {
            Some(c) => c,
            None => bail!("SetupUI instance not found"),
        };

        setup_ui.update(&mut self.terminal).await?;

        Ok(())
    }
}
