use std::cmp;

use anyhow::{Context, Result};
use tui::{
    layout::{Constraint, Direction, Layout},
    style::{Color, Style},
    widgets::{Block, Borders, Paragraph},
};

use self::widgets::{command_monitor, messages, room_info, rooms, status};
use super::UI;
use crate::app::status::Status;

pub mod widgets;

impl UI<'_> {
    pub async fn update(&mut self, status: &Status) -> Result<()> {
        let chunks = Layout::default()
            .direction(Direction::Vertical)
            .constraints([Constraint::Min(10), Constraint::Length(3)].as_ref())
            .split(self.terminal.size()?);

        let bottom_chunks = Layout::default()
            .direction(Direction::Horizontal)
            .constraints(
                [
                    Constraint::Length((status.state().to_string().len() + 2) as u16),
                    Constraint::Min(16),
                ]
                .as_ref(),
            )
            .split(chunks[1]);

        let main_chunks = Layout::default()
            .direction(Direction::Horizontal)
            .constraints(
                [
                    Constraint::Length(32),
                    Constraint::Min(16),
                    Constraint::Length(32),
                ]
                .as_ref(),
            )
            .split(chunks[0]);

        let left_chunks = Layout::default()
            .direction(Direction::Vertical)
            .constraints([Constraint::Length(5), Constraint::Min(4)].as_ref())
            .split(main_chunks[0]);

        let middle_chunks = Layout::default()
            .direction(Direction::Vertical)
            .constraints(
                [
                    Constraint::Min(4),
                    Constraint::Length(cmp::min(2 + self.message_compose.lines().len() as u16, 8)),
                ]
                .as_ref(),
            )
            .split(main_chunks[1]);

        let right_chunks = Layout::default()
            .direction(Direction::Vertical)
            .constraints([Constraint::Length(5), Constraint::Min(4)].as_ref())
            .split(main_chunks[2]);

        let colors = self
            .input_position
            .colors(&mut self.cli, &mut self.message_compose);

        // initiate the widgets
        let mode_indicator = Paragraph::new(status.state().to_string())
            .block(
                Block::default()
                    .borders(Borders::ALL)
                    .style(Style::default().fg(Color::DarkGray)),
            )
            .style(Style::default().fg(Color::LightYellow));
        let status_panel = status::init(status, &colors);
        let rooms_panel = rooms::init(status, &colors);
        let (messages_panel, mut messages_state) = messages::init(status.room(), &colors)
            .context("Failed to initiate the messages widget")?;
        let room_info_panel = room_info::init(status.room(), &colors);
        let command_monitor = command_monitor::init(status.status_messages(), &colors);

        // render the widgets
        self.terminal.draw(|frame| {
            frame.render_widget(status_panel, left_chunks[0]);
            frame.render_stateful_widget(rooms_panel, left_chunks[1], &mut self.rooms_state);
            frame.render_stateful_widget(messages_panel, middle_chunks[0], &mut messages_state);
            frame.render_widget(self.message_compose.widget(), middle_chunks[1]);
            frame.render_widget(mode_indicator, bottom_chunks[0]);
            match &self.cli {
                Some(cli) => frame.render_widget(cli.widget(), bottom_chunks[1]),
                None => (),
            };
            frame.render_widget(room_info_panel, right_chunks[0]);
            frame.render_widget(command_monitor, right_chunks[1]);
        })?;

        Ok(())
    }
}
