use anyhow::{Context, Result};
use matrix_sdk::ruma::events::{AnyMessageLikeEvent, AnyTimelineEvent};
use tui::{
    layout::Corner,
    style::{Color, Modifier, Style},
    text::{Span, Spans, Text},
    widgets::{Block, Borders, List, ListItem, ListState},
};

use crate::{app::status::Room, ui::central::InputPosition};

pub fn init<'a>(room: Option<&'a Room>, colors: &Vec<Color>) -> Result<(List<'a>, ListState)> {
    let content = match room {
        Some(room) => get_content_from_room(room).context("Failed to get content from room")?,
        None => vec![ListItem::new(Text::styled(
            "No room selected!",
            Style::default().fg(Color::Red),
        ))],
    };

    let mut messages_state = ListState::default();

    if let Some(room) = room {
        messages_state.select(room.view_scroll());
    }

    Ok((
        List::new(content)
            .block(
                Block::default()
                    .title("Messages")
                    .borders(Borders::ALL)
                    .style(Style::default().fg(colors[InputPosition::Messages as usize])),
            )
            .start_corner(Corner::BottomLeft)
            .highlight_symbol(">")
            .highlight_style(
                Style::default()
                    .fg(Color::LightMagenta)
                    .add_modifier(Modifier::BOLD),
            ),
        messages_state,
    ))
}

fn get_content_from_room(room: &Room) -> Result<Vec<ListItem>> {
    let results: Vec<Result<ListItem>> = room
        .timeline()
        .iter()
        .rev()
        .map(|event| filter_event(event).context("Failed to filter event"))
        .collect();

    let mut output = Vec::with_capacity(results.len());
    for result in results {
        output.push(result?);
    }
    Ok(output)
}

fn filter_event<'a>(event: &AnyTimelineEvent) -> Result<ListItem<'a>> {
    match event {
        // Message Like Events
        AnyTimelineEvent::MessageLike(message_like_event) => {
            let (content, color) = match &message_like_event {
                AnyMessageLikeEvent::RoomMessage(room_message_event) => {
                    let message_content = &room_message_event
                        .as_original()
                        .context("Failed to get inner original message_event")?
                        .content
                        .body();

                    (message_content.to_string(), Color::White)
                }
                _ => (
                    format!(
                        "~~~ not supported message like event: {} ~~~",
                        message_like_event.event_type().to_string()
                    ),
                    Color::Red,
                ),
            };
            let mut text = Text::styled(
                message_like_event.sender().to_string(),
                Style::default()
                    .fg(Color::Cyan)
                    .add_modifier(Modifier::BOLD),
            );
            text.extend(Text::styled(
                content.to_string(),
                Style::default().fg(color),
            ));
            Ok(ListItem::new(text))
        }

        // State Events
        AnyTimelineEvent::State(state) => Ok(ListItem::new(vec![Spans::from(vec![
            Span::styled(
                state.sender().to_string(),
                Style::default().fg(Color::DarkGray),
            ),
            Span::styled(": ", Style::default().fg(Color::DarkGray)),
            Span::styled(
                state.event_type().to_string(),
                Style::default().fg(Color::DarkGray),
            ),
        ])])),
    }
}
