use std::io::Stdout;

use anyhow::Result;
use tui::{
    backend::CrosstermBackend,
    layout::{Alignment, Constraint, Direction, Layout},
    style::{Color, Modifier, Style},
    text::Span,
    widgets::{Block, Borders, Paragraph},
    Terminal,
};
use tui_textarea::TextArea;

use crate::ui::{textarea_activate, textarea_inactivate};

pub struct UI<'a> {
    input_position: InputPosition,

    pub homeserver: TextArea<'a>,
    pub username: TextArea<'a>,
    pub password: TextArea<'a>,
    pub password_data: TextArea<'a>,
}

#[derive(Clone, Copy)]
pub enum InputPosition {
    Homeserver,
    Username,
    Password,
    Ok,
}

impl UI<'_> {
    pub fn new() -> Self {
        let mut homeserver = TextArea::new(vec!["https://matrix.org".to_string()]);
        let mut username = TextArea::default();
        let mut password = TextArea::default();
        let password_data = TextArea::default();

        homeserver.set_block(Block::default().title("Homeserver").borders(Borders::ALL));
        username.set_block(Block::default().title("Username").borders(Borders::ALL));
        password.set_block(Block::default().title("Password").borders(Borders::ALL));

        textarea_activate(&mut homeserver);
        textarea_inactivate(&mut username);
        textarea_inactivate(&mut password);

        Self {
            input_position: InputPosition::Homeserver,
            homeserver,
            username,
            password,
            password_data,
        }
    }

    pub fn cycle_input_position(&mut self) {
        self.input_position = match self.input_position {
            InputPosition::Homeserver => {
                textarea_inactivate(&mut self.homeserver);
                textarea_activate(&mut self.username);
                textarea_inactivate(&mut self.password);
                InputPosition::Username
            }
            InputPosition::Username => {
                textarea_inactivate(&mut self.homeserver);
                textarea_inactivate(&mut self.username);
                textarea_activate(&mut self.password);
                InputPosition::Password
            }
            InputPosition::Password => {
                textarea_inactivate(&mut self.homeserver);
                textarea_inactivate(&mut self.username);
                textarea_inactivate(&mut self.password);
                InputPosition::Ok
            }
            InputPosition::Ok => {
                textarea_activate(&mut self.homeserver);
                textarea_inactivate(&mut self.username);
                textarea_inactivate(&mut self.password);
                InputPosition::Homeserver
            }
        };
    }

    pub fn cycle_input_position_rev(&mut self) {
        self.input_position = match self.input_position {
            InputPosition::Homeserver => {
                textarea_inactivate(&mut self.homeserver);
                textarea_inactivate(&mut self.username);
                textarea_inactivate(&mut self.password);
                InputPosition::Ok
            }
            InputPosition::Username => {
                textarea_activate(&mut self.homeserver);
                textarea_inactivate(&mut self.username);
                textarea_inactivate(&mut self.password);
                InputPosition::Homeserver
            }
            InputPosition::Password => {
                textarea_inactivate(&mut self.homeserver);
                textarea_activate(&mut self.username);
                textarea_inactivate(&mut self.password);
                InputPosition::Username
            }
            InputPosition::Ok => {
                textarea_inactivate(&mut self.homeserver);
                textarea_inactivate(&mut self.username);
                textarea_activate(&mut self.password);
                InputPosition::Password
            }
        };
    }

    pub fn input_position(&self) -> &InputPosition {
        &self.input_position
    }

    pub async fn update(
        &mut self,
        terminal: &mut Terminal<CrosstermBackend<Stdout>>,
    ) -> Result<()> {
        let strings: Vec<String> = vec!["".to_owned(); 3];

        let content_ok = match self.input_position {
            InputPosition::Ok => {
                Span::styled("OK", Style::default().add_modifier(Modifier::UNDERLINED))
            }
            _ => Span::styled("OK", Style::default().fg(Color::DarkGray)),
        };

        let block = Block::default().title("Login").borders(Borders::ALL);

        let ok = Paragraph::new(content_ok).alignment(Alignment::Center);

        // define a 32 * 6 chunk in the middle of the screen
        let mut chunk = terminal.size()?;
        chunk.x = (chunk.width / 2) - 16;
        chunk.y = (chunk.height / 2) - 5;
        chunk.height = 12;
        chunk.width = 32;

        let mut split_chunk = chunk.clone();
        split_chunk.x += 1;
        split_chunk.y += 1;
        split_chunk.height -= 1;
        split_chunk.width -= 2;

        let chunks = Layout::default()
            .direction(Direction::Vertical)
            .constraints(
                [
                    Constraint::Length(3), // 0. Homserver:
                    Constraint::Length(3), // 1. Username:
                    Constraint::Length(3), // 2. Password:
                    Constraint::Length(1), // 3.    OK
                ]
                .as_ref(),
            )
            .split(split_chunk);

        terminal.draw(|frame| {
            frame.render_widget(block.clone(), chunk);
            frame.render_widget(self.homeserver.widget(), chunks[0]);
            frame.render_widget(self.username.widget(), chunks[1]);
            frame.render_widget(self.password.widget(), chunks[2]);
            frame.render_widget(ok.clone(), chunks[3]);
        })?;

        Ok(())
    }
}
